@@ -17,12 +17,20 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#define _XOPEN_SOURCE
+#define _XOPEN_SOURCE 500
 
-#include <stdio.h>
+#include <errno.h>
 #include <gio/gio.h>
-#include <stdlib.h>
+#include <glib/gstdio.h>
 #include <polkit/polkit.h>
+#include <stdio.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <sys/stat.h>
+#include <sys/types.h>
+#include <unistd.h>
+
 #include <libwhoopsie/identifier.h>
 
 #include "libwhoopsie-preferences.h"
@@ -34,6 +42,61 @@ static guint loop_shutdown = 0;
 static GKeyFile* key_file = NULL;
 static PolkitAuthority* authority = NULL;
 
+/* Hack for Ubuntu phone: check if path is an existing symlink to
+ * /etc/writable; if it is, update that instead */
+static const char* writable_filename (const char *config) {
+        const char *result = config;
+        struct stat sb;
+        char *linkname = NULL;
+        char *dirname = NULL;
+        char *resolved_path = NULL;
+        ssize_t r;
+        static char realfile_buf[PATH_MAX];
+
+        g_assert (config);
+        g_assert (g_path_is_absolute (config));
+
+        if (lstat (config, &sb) == -1) {
+            goto out;
+        }
+
+        linkname = g_malloc (sb.st_size + 1);
+        if (linkname == NULL) {
+            goto out;
+        }
+
+        r = readlink (config, linkname, sb.st_size + 1);
+
+        if (r < 0 || r > sb.st_size) {
+            goto out;
+        }
+
+        linkname[sb.st_size] = '\0';
+
+        /* Make absolute if relative */
+        if (!g_path_is_absolute (linkname)) {
+            dirname = g_path_get_dirname (config);
+            resolved_path = g_build_filename (
+                    dirname,
+                    linkname,
+                    NULL);
+            g_free (linkname);
+            linkname = resolved_path;
+        }
+
+        if (g_str_has_prefix (linkname, "/etc/writable")) {
+            g_snprintf (realfile_buf, sizeof (realfile_buf), "%s", linkname);
+            result = realfile_buf;
+        }
+
+out:
+        g_debug ("Writable path for %s is %s", config, result);
+
+        g_free (linkname);
+        g_free (dirname);
+        return result;
+}
+
 gboolean
 whoopsie_preferences_load_configuration (void)
 {
@@ -42,7 +105,8 @@ whoopsie_preferences_load_configuration (void)
     gboolean ret = TRUE;
     GError* error = NULL;
 
-    if (g_key_file_load_from_file (key_file, CONFIG_PATH,
+    if (g_key_file_load_from_file (key_file,
+                                   writable_filename(CONFIG_PATH),
                                    G_KEY_FILE_KEEP_COMMENTS, NULL)) {
         return TRUE;
     }
@@ -54,7 +118,7 @@ whoopsie_preferences_load_configuration (void)
         ret = FALSE;
         goto out;
     }
-    if (!g_file_set_contents (CONFIG_PATH, data, data_size, &error)) {
+    if (!g_file_set_contents (writable_filename(CONFIG_PATH), data, data_size, &error)) {
         g_print ("Could not write configuration: %s\n", error->message);
         ret = FALSE;
         goto out;
@@ -68,6 +132,11 @@ out:
     return ret;
 }
 
+static gboolean
+is_systemd() {
+    return g_file_test ("/run/systemd/system", G_FILE_TEST_IS_DIR);
+}
+
 static gboolean
 whoopsie_preferences_whoopsie_enabled (GError **error) {
     char* systemd_check_command[] = { "/bin/systemctl", "-q", "is-enabled", "whoopsie.service", NULL };
@@ -76,7 +145,7 @@ whoopsie_preferences_whoopsie_enabled (GError **error) {
     int exit_code = 0;
 
     /* systemd running */
-    if (g_file_test ("/run/systemd/system", G_FILE_TEST_IS_DIR)) {
+    if (is_systemd()) {
         if (g_spawn_sync (NULL, systemd_check_command, NULL, 0, NULL, NULL, NULL, NULL, &exit_code, error))
             report_crashes = (exit_code == 0);
     }
@@ -243,9 +312,14 @@ whoopsie_preferences_changed (WhoopsiePreferences* object, GParamSpec* pspec,
     gboolean saved_value, new_value;
     GError* error = NULL;
     char* data;
+    char* errstr;
     gsize data_size;
     gboolean crashes = !g_strcmp0(user_data, "report_crashes");
     char* crash_enable_command[] = { "/usr/sbin/update-rc.d", "whoopsie", NULL, NULL };
+    char* initctl_reload_command[] = { "/sbin/initctl", "reload-configuration", NULL };
+    gboolean spawn_ok;
+    gint exit_status;
+    FILE* f = NULL;
 
     increase_shutdown_timeout ();
 
@@ -269,10 +343,35 @@ whoopsie_preferences_changed (WhoopsiePreferences* object, GParamSpec* pspec,
         error = NULL;
 
         if (crashes) {
-            if (!g_spawn_sync (NULL, crash_enable_command, NULL, 0, NULL, NULL, NULL, NULL, NULL, &error)) {
-                g_print ("Could not run invoke-rc.d: %s\n", error->message);
-                g_error_free (error);
-                return FALSE;
+            spawn_ok = g_spawn_sync (NULL, crash_enable_command, NULL, 0, NULL,
+                            NULL, NULL, NULL, &exit_status, &error);
+            if (!spawn_ok || exit_status != 0) {
+                if (error) {
+                    errstr = g_strdup (error->message);
+                    g_error_free (error);
+                } else
+                    errstr = g_strdup_printf ("exited with code %d", exit_status);
+                g_print ("Could not run update-rc.d: %s\n", errstr);
+                g_free (errstr);
+
+                /* It's a real error for systemd */
+                if (is_systemd())
+                    return FALSE;
+
+                /* Works for enabling (truncating) on upstart only atm */
+                if (new_value) {
+                    g_debug ("Truncating the override file manually");
+                    f = fopen ("/etc/init/whoopsie.override", "w");
+                    if (f) {
+                        fclose (f);
+			if (!g_spawn_sync (NULL, initctl_reload_command, NULL, 0, NULL,
+					NULL, NULL, NULL, &exit_status, &error))
+				g_print ("Could not run initctl reload-configuration: %s\n", errstr);
+		    } else {
+                        g_warning ("Couldn't open override file: %s", strerror(errno));
+                        return FALSE;
+                    }
+                }
             }
         } else {
             g_key_file_set_boolean (key_file, "General", user_data, new_value);
@@ -281,7 +380,7 @@ whoopsie_preferences_changed (WhoopsiePreferences* object, GParamSpec* pspec,
                 g_print ("Could not process configuration: %s\n", error->message);
                 return FALSE;
             }
-            if (!g_file_set_contents (CONFIG_PATH, data, data_size, &error)) {
+            if (!g_file_set_contents (writable_filename(CONFIG_PATH), data, data_size, &error)) {
                 g_print ("Could not write configuration: %s\n", error->message);
                 return FALSE;
             }
